/*
    Copyright (C) 2004 Adam Olsen

	This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 1, or (at your option)
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <gjournal/journalwindow.hpp>

namespace gjournal
{
	JournalWindow::JournalWindow( const char *loadFile )
	: dateLabel( "Entry Date: " )
	{
		jf = new JournalFile( this );
		_checkForTemps();
		_checkForOldFileFormat();

		// set the icon for the window
		set_icon( Gdk::Pixbuf::create_from_xpm_data( gtk_journal_icon_xpm ) );

		// layout the child components
		_layoutComponents();

		// set up the various signal handlers
		signal_delete_event().connect( SigC::slot( *this, &JournalWindow::_closeHandler ) );
		calendar.signal_day_selected_double_click().connect( SigC::slot( *this, &JournalWindow::_daySelectHandler ) );
		modified = journalEntry.get_buffer()->signal_changed().connect( SigC::slot( *this, &JournalWindow::_journalEntryChangeHandler ) );

		_newFileHandler();
		loading = false;
		if( loadFile )
		{
			loading = true;
			_loadFileHandler( loadFile );
		}

		show_all_children();
		show();
	}

	void JournalWindow::_checkForOldFileFormat()
	{
		std::string oldJournal = Glib::get_home_dir();
		oldJournal += Globals::separator;
		oldJournal += ".gtkjournal";
		oldJournal += Globals::separator;
		oldJournal += "journal.enc";

		if( Glib::file_test( oldJournal, Glib::FILE_TEST_EXISTS ) )
		{
			Glib::ustring mess = "It looks like you haven't yet converted your journal file from the old format yet. ";
			mess += "You will need to do this before you can access your old journal.  ";
			mess += "You can find information on how in the sourcecode's README file.  Please remove your \"~/.gtkjournal\" directory ";
			mess += "if you have already converted your old journal file and you don't want to see this message again.";
			Gtk::MessageDialog dialog( mess, Gtk::MESSAGE_INFO );
			dialog.run();
			dialog.hide();
		}
	}

	void JournalWindow::_loadFileHandler( const char *f )
	{
		std::string oldTempFile = tempFile;
		std::string error;
		std::string file( f );

		aolib::LoginDialog login( "Please enter the username and password" );
		bool check = false;

		if( !Glib::file_test( file, Glib::FILE_TEST_EXISTS ) )
		{
			Gtk::MessageDialog dialog( "Sorry, the file you specified does not exist.", Gtk::MESSAGE_WARNING );
			dialog.run();
			dialog.hide();
			return;
		}

		sql.disconnect();

		int result;
		tempFile = jf->_getNewTempFile();
		aolib::BFCryptFile crypt( "", "" );

		while( !check )
		{
			result = login.run();
			if( result != Gtk::RESPONSE_OK )
			{
				if( loading )
				{
					_shredTempFile( oldTempFile );
					exit( -1 );
				}

				check = true;
				tempFile = oldTempFile;
			}
			else {
				crypt.setKeys( login.getUsername(), login.getPassword() );
				if( !crypt.decryptFile( file, tempFile, error ) )
				{
					Gtk::MessageDialog dialog( error, Gtk::MESSAGE_WARNING );
					dialog.run();
					dialog.hide();
				}
				else check = true;
			}
		}

		if( result != Gtk::RESPONSE_OK )
		{
			tempFile = oldTempFile;
			_dbConnect();
			return;
		}

		if( Globals::plainTextPassword ) delete Globals::plainTextPassword;
		if( Globals::username ) delete Globals::username;
		Globals::plainTextPassword = new std::string( login.getPassword() );
		Globals::username = new std::string( login.getUsername() );

		loading = false;
		journalFile = file;
		_shredTempFile( oldTempFile );
		_dbConnect();
		_todayHandler();
		std::string title = "Gtk+ Journal - ";
		if( journalFile != "" )	title += basename( journalFile.c_str() );
		else title += "New Journal";
		set_title( title );
	}

	void JournalWindow::_newFileHandler()
	{
		if( !savedFlag )
		{
			Gtk::MessageDialog dialog( "Save your current Journal?", Gtk::MESSAGE_QUESTION, Gtk::BUTTONS_YES_NO );
			dialog.add_button( Gtk::Stock::CANCEL, Gtk::RESPONSE_CANCEL );
			int result = dialog.run();
			dialog.hide();
			if( result == Gtk::RESPONSE_CANCEL ) return;
			if( result == Gtk::RESPONSE_YES )
			{
				_saveHandler();
				if( !savedFlag ) return;
			}
		}

		if( tempFile != "" ) _shredTempFile( tempFile );
		if( sql.isConnected() ) sql.disconnect();

		tempFile = jf->_getNewTempFile();

		std::string error;
		if( !jf->_createNewFile( tempFile, error ) )
		{
			Gtk::MessageDialog dialog( "Could not create temp file.\n" + error, Gtk::MESSAGE_ERROR );
			dialog.run();
			dialog.hide();
			if( !Globals::plainTextPassword ) exit( -1 );
			return;
		}
		else {
			Glib::RefPtr<Gtk::TextBuffer> buf = journalEntry.get_buffer();
			buf->set_text( "" );
			journalEntry.set_buffer( buf );
			journalFile = "";
			_dbConnect();
			_todayHandler();
		}

		if( Globals::username ) delete Globals::username;
		Globals::username = 0;
		if( Globals::plainTextPassword ) delete Globals::plainTextPassword;
		Globals::plainTextPassword = 0;
		statusBar.pop();
		statusBar.push( "Status: Unmodified" );
		std::string title = "Gtk+ Journal - *New Journal";
		set_title( title );
		savedFlag = true;
	}

	void JournalWindow::_daySelectHandler()
	{
		guint month, day, year;

		// get the day that is currently selected
		calendar.get_date( year, month, day );

		// and load the entry for it
		_loadEntry( ( month + 1 ), day, year );
	}

	void JournalWindow::_openHandler()
	{
		if( !savedFlag )
		{
			// if the file has not yet been saved yet, find out if they want to save it
			Gtk::MessageDialog dialog( "Save your current Journal?", Gtk::MESSAGE_QUESTION, Gtk::BUTTONS_YES_NO );
			dialog.add_button( Gtk::Stock::CANCEL, Gtk::RESPONSE_CANCEL );
			int result = dialog.run();
			dialog.hide();
			if( result == Gtk::RESPONSE_CANCEL ) return;
			if( result == Gtk::RESPONSE_YES )
			{
				// save the file
				_saveHandler();
				if( !savedFlag ) return;
			}
		}

		// show a file selection box
		Gtk::FileSelection fs( "Open Journal File" );
		fs.set_transient_for( *this );
		int result = fs.run();
		if( result == Gtk::RESPONSE_OK )
		{
			fs.hide();
			_loadFileHandler( fs.get_filename().c_str() );
		}
	}

	void JournalWindow::_layoutComponents()
	{
		add( mainBox );
		mainBox.pack_start( menuBar, Gtk::PACK_SHRINK );

		// File Menu stuff
		menuBar.items().push_back( Gtk::Menu_Helpers::MenuElem( "_File", fileMenu ) );
		fileMenu.items().push_back( Gtk::Menu_Helpers::StockMenuElem( Gtk::Stock::NEW, SigC::slot( *this, &JournalWindow::_newFileHandler ) ) );
		fileMenu.items().push_back( Gtk::Menu_Helpers::StockMenuElem( Gtk::Stock::OPEN, SigC::slot( *this, &JournalWindow::_openHandler ) ) );
		fileMenu.items().push_back( Gtk::Menu_Helpers::StockMenuElem( Gtk::Stock::SAVE, SigC::slot( *this, &JournalWindow::_saveHandler ) ) );
		fileMenu.items().push_back( Gtk::Menu_Helpers::StockMenuElem( Gtk::Stock::SAVE_AS, SigC::slot( *this, &JournalWindow::_saveAsHandler ) ) );

		fileMenu.items().push_back( Gtk::Menu_Helpers::SeparatorElem() );
		fileMenu.items().push_back( Gtk::Menu_Helpers::StockMenuElem( Gtk::Stock::QUIT, SigC::slot( *this, &JournalWindow::_exitHandler ) ) );

		// Edit Menu stuff
		menuBar.items().push_back( Gtk::Menu_Helpers::MenuElem( "_Edit", editMenu ) );
		editMenu.items().push_back( Gtk::Menu_Helpers::StockMenuElem( Gtk::Stock::PREFERENCES, SigC::slot( *this, &JournalWindow::_preferencesHandler ) ) );

		// Help Menu stuff
		menuBar.items().push_back( Gtk::Menu_Helpers::StockMenuElem( Gtk::Stock::HELP, helpMenu ) );
		helpMenu.items().push_back( Gtk::Menu_Helpers::MenuElem( "_About", SigC::slot( *this, &JournalWindow::_aboutHandler ) ) );

		// create a toolbar with three items
		mainBox.pack_start( toolBar, Gtk::PACK_SHRINK );
		{
			using namespace Gtk::Toolbar_Helpers;

			// create some buttons
			StockElem newItem( Gtk::Stock::NEW, SigC::slot( *this, &JournalWindow::_newFileHandler ), "New Journal" );
			StockElem open( Gtk::Stock::OPEN, SigC::slot( *this, &JournalWindow::_openHandler ), "Open Journal" );
			StockElem save( Gtk::Stock::SAVE, SigC::slot( *this, &JournalWindow::_saveHandler ), "Save Journal" );
			StockElem prefs( Gtk::Stock::PREFERENCES, SigC::slot( *this, &JournalWindow::_preferencesHandler ), "Preferences" );
			StockElem today( Gtk::Stock::DIALOG_INFO, SigC::slot( *this, &JournalWindow::_todayHandler ), "Today's Entry" );

			// add the buttons
			toolBar.tools().push_back( newItem );
			toolBar.tools().push_back( open );
			toolBar.tools().push_back( save );
			toolBar.tools().push_back( prefs );
			toolBar.tools().push_back( Space() );
			toolBar.tools().push_back( today );

			// add the date label
			Element date( dateLabel );
			toolBar.tools().push_back( date );
		}

		calendarBox.set_size_request( 210, 70 );

		mainBox.pack_start( centerBox );
		centerBox.pack_start( calendarBox, Gtk::PACK_SHRINK );
		calendarBox.pack_start( calendar, Gtk::PACK_SHRINK );
		window.add( journalEntry );
		journalEntry.set_wrap_mode( Gtk::WRAP_WORD );
		window.set_shadow_type( Gtk::SHADOW_ETCHED_IN );
		window.set_policy( Gtk::POLICY_AUTOMATIC, Gtk::POLICY_AUTOMATIC );
		centerBox.pack_start( window );

		mainBox.pack_start( statusBar, Gtk::PACK_SHRINK );

		set_default_size( 750, 480 );
		// center the journal window
		set_position( Gtk::WIN_POS_CENTER );
	}

	void JournalWindow::_loadEntry( int month, int mday, int year )
	{
		std::stringstream date;
		char *error;

		if( currentEntry != "" ) _saveCurrentEntry();

		// format the date in a string like 2004-01-05
		date << year << "-";
		date << std::setw( 2 ) << std::setfill( '0' ) << month << "-";
		date << std::setw( 2 ) << std::setfill( '0' ) << mday;
		std::string query = "SELECT Date, Entry FROM entries WHERE Date=\"" + date.str() + "\";";

		// we need to disconnect the modified signal handler because switching days
		// and clearing out the buffer will set it off
		modified.disconnect();

		// select for today from sqlite

		sqlitem::SQliteResult *res = sql.query( query );

		if( !res )
		{
			_sqlError( error );
		}
		else {
			currentEntry = date.str();
			date.clear();
			date.str( "" );
			date << aolib::long_month_from_int( month ) << " " << mday << ", " << year;

			// set the date label to the date being loaded
			dateLabel.set_text( " Entry Date: " + date.str() );

			// put focus into the entry box
			journalEntry.grab_focus();
			Glib::RefPtr<Gtk::TextBuffer> buf = journalEntry.get_buffer();

			// select the day specified in the calendar
			calendar.select_month( ( month - 1 ), year );
			calendar.select_day( mday );

			if( !res->numRows() ) buf->set_text( "" );
			else {
				std::map<std::string, std::string> row = res->next();
				buf->set_text( row["Entry"] );
			}

			res->finalize();
			journalEntry.set_buffer( buf );
		}

		// reconnect the modified listener
		modified = journalEntry.get_buffer()->signal_changed().connect( SigC::slot( *this, &JournalWindow::_journalEntryChangeHandler ) );
	}

	void JournalWindow::_journalEntryChangeHandler()
	{
		// mark the journal as not saved
		savedFlag = false;
		std::string title = "Gtk+ Journal - *";
		if( journalFile != "" )	title += basename( journalFile.c_str() );
		else title += "New Journal";
		set_title( title );
		statusBar.pop();
		statusBar.push( "Status: Modified (not saved)" );
	}

	void JournalWindow::_saveCurrentEntry()
	{
		if( journalEntry.get_buffer()->get_text() == "" ) return;

		std::string entry = sqlitem::SQliteManager::addSlashes( journalEntry.get_buffer()->get_text() );
		char *error;
		std::string query = "REPLACE INTO entries( Date, Entry ) VALUES( \"" + currentEntry + "\", \"" + entry + "\" );";

		// save the current entry into the database
		if( sql.go( query ) == -1 )
		{
			_sqlError( sql.lastError() );
		}
	}

	void JournalWindow::_dbConnect()
	{
		// connect to the SQL database
		if( !sql.connect( tempFile ) )
		{
			_sqlError( sql.lastError(), true );
		}
	}

	void JournalWindow::_saveAsHandler()
	{
		if( !Globals::plainTextPassword )
		{
			jf->_preferences( true );
			return;
		}

		bool check = false;
		Gtk::FileSelection fs( "Save As" );
		fs.set_transient_for( *this );

		while( !check )
		{
			int result = fs.run();
			if( result != Gtk::RESPONSE_OK ) return;

			if( Glib::file_test( fs.get_filename(), Glib::FILE_TEST_EXISTS ) )
			{
				Gtk::MessageDialog dialog( fs, "The filename you have chosen already exists.  Do you want to overrwite it?", Gtk::MESSAGE_QUESTION,
					Gtk::BUTTONS_YES_NO );
				int res = dialog.run();
				if( res == Gtk::RESPONSE_YES ) check = true;
			}
			else check = true;
		}

		journalFile = fs.get_filename();
		_saveHandler();
	}

	void JournalWindow::_saveHandler()
	{
		std::string error;

		if( currentEntry != "" ) _saveCurrentEntry();

		if( journalFile == "" )
		{
			_saveAsHandler();
			return;
		}

		sql.disconnect();
		Gtk::MessageDialog pd( "Encrypting Journal, please wait...", Gtk::MESSAGE_INFO, Gtk::BUTTONS_NONE );
		pd.show();
		while( Gtk::Main::events_pending() )
		{
			Gtk::Main::iteration();
		}
		// encrypt the temporary database to the journalFile
		aolib::BFCryptFile crypt( *Globals::username, *Globals::plainTextPassword );

		if( !crypt.encryptFile( tempFile, journalFile, error ) )
		{
			Gtk::MessageDialog dialog( *this, error, Gtk::MESSAGE_ERROR );
			dialog.run();
		}
		else {
			savedFlag = true;
			std::string title = "Gtk+ Journal - ";
			if( journalFile != "" )	title += basename( journalFile.c_str() );
			else title += "New Journal";
			set_title( title );
			statusBar.pop();
			statusBar.push( "Status: Unmodified (saved)" );
		}

		pd.hide();

		// and then reconned to the temp database
		_dbConnect();
	}

	void JournalWindow::_todayHandler()
	{
		// get the current date
		time_t t = time( NULL );
		tm *the = localtime( &t );

		// load the entry from the current date
		_loadEntry( ( the->tm_mon + 1 ), the->tm_mday, ( the->tm_year + 1900 ) );
	}

	void JournalWindow::_preferencesHandler()
	{
		jf->_preferences();
	}

	void JournalWindow::_sqlError( std::string error, bool e )
	{
		if( error == "" ) error = "Unknown SQL error";

		// display the error
		Gtk::MessageDialog dialog( *this, "There has been an error executing your request.\n" + error, Gtk::MESSAGE_ERROR );
		dialog.run();

		// if it's a fatal error, bail out
		if( e ) exit( -1 );
	}

	void JournalWindow::_exitHandler()
	{
		int result = Gtk::RESPONSE_NO;

		// if they haven't saved, ask them if they really want to exit
		if( !savedFlag )
		{
			// find out if they want to quit
			Gtk::MessageDialog dialog( *this, "Would you like to save your Journal before quitting?",
				Gtk::MESSAGE_QUESTION, Gtk::BUTTONS_YES_NO );
			dialog.add_button( Gtk::Stock::CANCEL, Gtk::RESPONSE_CANCEL );

			result = dialog.run();
			dialog.hide();
			if( result == Gtk::RESPONSE_CANCEL ) return;
			if( result == Gtk::RESPONSE_YES )
			{
				_saveHandler();
				if( !savedFlag ) return;
			}
		}

		sql.disconnect(); // close the database connection
		hide();

		_shredTempFile( tempFile );

		g_print( "Goodbye!\n" );
		exit( 0 );
	}

	void JournalWindow::_shredTempFile( std::string file )
	{
		Gtk::MessageDialog dialog( "Shredding temp file, please wait...", Gtk::MESSAGE_INFO, Gtk::BUTTONS_NONE );
		dialog.show();

		while( Gtk::Main::events_pending() )
		{
			Gtk::Main::iteration();
		}

		dialog.hide();

		aolib::BFCryptFile::shredFile( file.c_str() );
	}

	void JournalWindow::_checkForTemps()
	{
		DIR *dir = opendir( Globals::TEMP_DIR.c_str() );
		dirent *d;
		bool warned = false;
		while( ( d = readdir( dir ) ) != 0 )
		{
			std::string filename = d->d_name;
			int b = filename.find( ".tmp___gtkj___", 0 );
			if( b != std::string::npos )
			{
				if( !warned )
				{
					Glib::ustring message = "It appears as though Gtk+ Journal was exiteed abnormally.  You should always try to exit ";
					message += "correctly by choosing 'File->Quit' or by typing CTRL+Q, because not doing so can leave your data unprotected.";

					Gtk::MessageDialog dialog( *this, message );
					dialog.run();
					dialog.hide();
					warned = true;
				}
				_shredTempFile( Globals::TEMP_DIR + filename );
			}
		}
		closedir( dir );
	}

	bool JournalWindow::_closeHandler( GdkEventAny *event )
	{
		// basically just call the exit handler
		_exitHandler();
		return true;
	}

	void JournalWindow::_aboutHandler()
	{
		Gtk::MessageDialog about( *this, "Gtk+ Journal " + Globals::version + "\nBy: Adam R Olsen (adamolsen@net-bizz.com)\n(c)2004" );
		about.run();
	}

	JournalWindow::~JournalWindow()
	{
		delete jf;
		delete Globals::plainTextPassword;
		delete Globals::username;
	}
}
